%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% K.Hirai added at 2008/01/08 ('judgementOfConvergence' and 'calculatingFocalLengthMode')
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function ThreeDRec(file1 , file2  , outputpathInitF , outputpathF , outputpathT , outputpathR , outputpath3D , ...
                   outputpathFocal1 , outputpathFocal2 , ...
                   initializingMethodName, ...
                   calculatingFocalLengthMode)


default_focal_length = 600.0


% <start data input>
disp('------------------------------read data------------------------------')
disp('')
[list1 , length1] = readlines(file1);
[list1 , length1] = scanNumberTuples(2 , list1 , length1);

[list2 , length2] = readlines(file2);
[list2 , length2] = scanNumberTuples(2 , list2 , length2);


if length1 ~= length2 , error('the two files contain different numbers of data.'); end


pairList = cell(length1 , 1 , 4);


for i = 1 : length1
    pairList{i , 1}{1} = list1{i , 1}{1};
    pairList{i , 1}{2} = list1{i , 1}{2};
    pairList{i , 1}{3} = list2{i , 1}{1};
    pairList{i , 1}{4} = list2{i , 1}{2};
end

% 入力の元の値を保持
pairListOriginal = pairList;

% <end data input>


% <start initializing>
disp('------------------------------calculating start------------------------------')
disp('')
disp('------------------------------initialize------------------------------')

if strcmpi(initializingMethodName , 'LSM') , 
    eigenVector = initializeWithLSM(pairList , length1 , default_focal_length);
elseif strcmpi(initializingMethodName , 'Taubin') , 
    eigenVector = initializeWithTaubin(pairList , length1 , default_focal_length);
else  
    error(['unknown initilizing method:' initialinzingMethodName ]);
end


F = [eigenVector(1) , eigenVector(2) , eigenVector(3) ;
     eigenVector(4) , eigenVector(5) , eigenVector(6) ;
     eigenVector(7) , eigenVector(8) , eigenVector(9)];
F = F / norm(F , 'fro');
initF = F

% <end initializing>


% <begin main program>

eigenVector = eigenVector / norm(eigenVector);

try
    %% added flag into each function arguments by K.Hirai
    %(2008/01/12)
    disp('')
    disp(['--------------------main program start--------------------------'])
    [F , pairList] = main_program(pairList , length1 , eigenVector , ...
                                  default_focal_length, 'update');
    xFx = 0.0;
    for i = 1 : length1
        x1 = list1{i}{1} / default_focal_length; 
        y1 = list1{i}{2} / default_focal_length;
        x2 = list2{i}{1} / default_focal_length; 
        y2 = list2{i}{2} / default_focal_length;
        xFx = xFx + abs(dot([x1 ; y1; 1.00] , F * [x2 ; y2 ; 1.00]));
    end
    
    disp([''])


    %% 焦点距離の計算 
    disp('------------------------------calc focal Length------------------------------')
    [focalLength1 , focalLength2] = calcFocalLength(F , default_focal_length, calculatingFocalLengthMode)
    disp('')

    %% カメラ 1 の焦点距離を出力
    lines = serializeNumberTuples(1 , {{focalLength1}} ,  1 , '   ');
    writeLines(outputpathFocal1  , lines , 1);

    %% カメラ 2 の焦点距離を出力
    lines = serializeNumberTuples(1 , {{focalLength2}} ,  1 , '   ');
    writeLines(outputpathFocal2  , lines , 1);

catch 
    disp(lasterr)
    disp('ERROR: calc focal length')
    focalLength1 = default_focal_length;
    focalLength2 = default_focal_length;
    
    writeLines(outputpathFocal1  , {'NULL'} , 1);
    writeLines(outputpathFocal2  , {'NULL'} , 1);
end

disp('---------------------calc motion parameter---------------------------')
[ t , R ] = calcMotionParameter(pairList , length1 , F , default_focal_length , focalLength1 , focalLength2 )
disp('')


disp('---------------------re-calc fundamental matrix---------------------------')
F = createFundamentalMatrix(default_focal_length , focalLength1 , focalLength2 , R ,t )
disp('')


disp('---------------------optical correction method---------------------------')
vectorU = zeros(9,1);
vectorU(1) = F(1,1);
vectorU(2) = F(1,2);
vectorU(3) = F(1,3);
vectorU(4) = F(2,1);
vectorU(5) = F(2,2);
vectorU(6) = F(2,3);
vectorU(7) = F(3,1);
vectorU(8) = F(3,2);
vectorU(9) = F(3,3);
[F , pairList] = main_program(pairListOriginal, length1, vectorU, default_focal_length, 'not-update');
disp('')


disp('-------------------------calc 3D position----------------------------')
positionList = calc3DPositionList(pairList , length1 , default_focal_length , focalLength1 , focalLength2 , t , R);
disp([''])


%% 基礎行列の初期値を出力
fList = cell(1 , 3);
fList{1} = {initF(1 , 1) , initF(1 , 2) , initF(1 , 3)};
fList{2} = {initF(2 , 1) , initF(2 , 2) , initF(2 , 3)};
fList{3} = {initF(3 , 1) , initF(3 , 2) , initF(3 , 3)};
lines = serializeNumberTuples(3 , fList ,  3 , '   ');
writeLines(outputpathInitF  , lines , 3);

%% 最終的な基礎行列を出力
fList{1} = {F(1 , 1) , F(1 , 2) , F(1 , 3)};
fList{2} = {F(2 , 1) , F(2 , 2) , F(2 , 3)};
fList{3} = {F(3 , 1) , F(3 , 2) , F(3 , 3)};
lines = serializeNumberTuples(3 , fList ,  3 , '   ');
writeLines(outputpathF  , lines , 3);

%% カメラ間の並進ベクトルを出力
tList = {{t(1) , t(2) , t(3)}};
lines = serializeNumberTuples(3 , tList ,  1 , '   ');
writeLines(outputpathT  , lines , 1);

%% カメラ間の回転行列を出力
rList = cell(1 , 3);
rList{1} = {R(1 , 1) , R(1 , 2) , R(1 , 3)};
rList{2} = {R(2 , 1) , R(2 , 2) , R(2 , 3)};
rList{3} = {R(3 , 1) , R(3 , 2) , R(3 , 3)};
lines = serializeNumberTuples(3 , rList , 3 , '   ');
writeLines(outputpathR  , lines , 3);

%% 3 次元位置を出力
lines = serializeNumberTuples(3 , positionList , length1 , '   ');
write3DPositionForDisplay3D(outputpath3D  , lines , length1);
